
module FTree where

import Mpi

-- (1) Datatype definition -----------------------------------------------------

data FTree a c = Unit c | Comp (a, (FTree a c, FTree a c)) deriving Show

inFTree = either Unit Comp

outFTree :: FTree a c -> Either c (a,(FTree a c,FTree a c))
outFTree (Unit c)           = Left c
outFTree (Comp (a,(t1,t2))) = Right(a,(t1,t2))

-- (2) Ana + cata + hylo -------------------------------------------------------

cataFTree a = a . (recFTree (cataFTree a)) . outFTree

anaFTree f = inFTree . (recFTree (anaFTree f) ) . f

hyloFTree a c = cataFTree a . anaFTree c

recFTree f = id -|- (id >< (f >< f))

-- (3) Map ---------------------------------------------------------------------

instance BiFunctor FTree
         where bmap f g = cataFTree ( inFTree . (g -|- (f >< (id >< id))) )


