-- (c) M.P.I 98/99 -- Date: 98.12.03

module LTree where

import Mpi

-- (1) Datatype definition -----------------------------------------------------

data LTree a = Leaf a | Split (LTree a, LTree a) deriving Show

inLTree = either Leaf Split

outLTree :: LTree a -> Either a (LTree a,LTree a)
outLTree (Leaf a)     = i1   a
outLTree (Split (t1,t2)) = i2    (t1,t2)

-- (2) Ana + cata + hylo -------------------------------------------------------

cataLTree a = a . (recLTree (cataLTree a)) . outLTree

anaLTree f = inLTree . (recLTree (anaLTree f) ) . f

hyloLTree a c = cataLTree a . anaLTree c

recLTree f = id -|- (f >< f)

-- (3) Map ---------------------------------------------------------------------

instance Functor LTree
         where map f = cataLTree ( inLTree . ( f -|- (id >< id)))

-- (4) Examples ----------------------------------------------------------------

-- (4.1) Flatten ---------------------------------------------------------------

flatten = cataLTree (either sing conc)

-- where

sing   x = [x]
conc(l,r)= l ++ r

-- (4.2) Double factorial ------------------------------------------------------

dfac 0 = 1
dfac n = (hyloLTree (either id mul) interval) (1,n)

-- where

mul(x,y)=x*y

interval(n,m) | n==m      = i1   n
              | otherwise = i2   ((n,k),(k+1,m))
                            where k = div (n+m) 2

-- (4.3) Fibonacci -------------------------------------------------------------

fib =  hyloLTree (either id add) getN

-- where

add(x,y)=x+y

getN n | n < 2     = i1   1
       | otherwise = i2   (n-1,n-2)

-- (4.4) Merge sort ------------------------------------------------------------

mSort :: Ord a => [a] -> [a]
mSort [] = []
mSort l = hyloLTree (either sing merge) lsplit l

--where

merge (l,[])                  = l
merge ([],r)                  = r
merge (x:xs,y:ys) | x < y     = x : merge(xs,y:ys) 
                  | otherwise = y : merge(x:xs,ys)

lsplit [x] = i1   x
lsplit l   = i2    (sep l)

sep []    = ([],[])
sep (x:t) = let (l,r) = sep t in (x:r,l)  -- a List cata

--------------------------------------------------------------------------------

